/*
/ Projektname : LCD-I2C-Modul
/ Dateiname   : I2C.c
/ Version     : 1.0.0
/ Datum       : 18.01.2014
/ Compiler    : WinAVR 20100110 libc 1_7_1
/ Autor       : Johannes Kretzmann
/ Copyright   : johannes.kretzmann@projektsammlung.de
/               www.projektsammlung.de
*/

 
#include <avr/io.h>
#include <inttypes.h>
#include "I2C.h"


/* -------------------------------------------------------------------------------------
   Initialisierungsroutine fr den I2C-Bus 
                        In dieser Funktion werden die Parameter fr die Portpins so eingestellt,
                        dass ber sie per Software der I2C-Bus abgebildet werden kann. Dazu werden die beiden Ports
                        die an den Leitungen fr SCL und SDA angeschlossen werden als Eingnge definiert und die 
                        internen Pull-Up-Widerstnde aktiviert, so dass durch diese Paramentrierung hochohmige Eingangsports
                        entstehen. Beide Leitungen werden auf High-Pegel gesetzt, das sich in diesem Zustand der I2C-Bus
                        im Ruhezustande befindet.
   Funktion :bergabe:   Input     : -
                        Return    : -
--------------------------------------------------------------------------------------*/ 
void Init_I2C(void)
{
    PORTB &= ~((1<<I2C_SCL) | (1<<I2C_SDA));               // Pull-Up-Widerstnde der Port-Pins fr SCL und SDA deaktivieren;
    DDRB  &= ~((1<<I2C_SCL) | (1<<I2C_SDA));               // Port-Pins von SCL und SDA als Eingnge definieren; High-Pegel; SCL = 1, SDA = 1
                                                             // Signal geht auf High-Pegel durch die externen Pull-Up-Widerstnde 
    return;
}



/* -------------------------------------------------------------------------------------
   Funktion :           I2C-Bus-Startbedingung: Diese Funktion erzeugt die Pegel auf den beiden
                        Leitungen die fr die Startbedingung des I2C-Bus bemtigt werden.
   Parameterbergabe:   Input     : -
                        Return    : -
--------------------------------------------------------------------------------------*/
void I2C_Start(void)
{
    DDRB  &= ~((1<<I2C_SDA) | (1<<I2C_SCL));               // Port-Pins von SCL und SDA als Eingnge definieren;
                                                             // Signal geht auf High Pegel durch die Pull-Up-Widerstnde 
    DDRB  |= (1<<I2C_SDA);                                  // Port-Pin von SDA als Ausgang definieren; Low-Pegel; SDA = 0;
    asm volatile ("nop");                                    // notwendige Verzgerung
    asm volatile ("nop");                                    // notwendige Verzgerung
    DDRB  |= (1<<I2C_SCL);                                  // Port-Pin von SCL als Ausgang definieren; Low-Pegel; SCL = 0

    return;
}


/* -------------------------------------------------------------------------------------
   Funktion :           Diese Funktion empfngt die ber den I2C-Bus gesendeten Daten-Bits.
                        Hierzu werden die fr das Einlesen bentigten Pegelzusnde auf den beide
                        Leitungen erzeugt und die Leitungszustnde ausgewertet, bis alle 8 Bits
                        eingelesen sind (Das MSB ist das zuerst eingelesene Bit)
   Parameterbergabe:   Input     : -
                        Return    : data       Byte das ber den I2C-Bus eingelesen wurde
--------------------------------------------------------------------------------------*/
uint8_t I2C_Receive(void)
{

    uint8_t data=0;
    uint8_t I2C_Daten_Bit_Zaehle=0b10000000;

    while(I2C_Daten_Bit_Zaehle>0)
    {

        DDRB  &= ~(1<<I2C_SCL);                             // Port-Pin von SCL als Eingang definieren; High-Pegel; SCL = 1 ; durch Pull-Up

        if (PIN_I2C_SDA & (1<<BIT_I2C_SDA))
            data |= I2C_Daten_Bit_Zaehle;
        else
            data &= ~I2C_Daten_Bit_Zaehle;

        DDRB  |= (1<<I2C_SCL);                              // Port-Pin von SCL als Ausgang definieren; Low-Pegel;  SCL = 0;

        I2C_Daten_Bit_Zaehle=(I2C_Daten_Bit_Zaehle>>1);
    }


    DDRB  &= ~(1<<I2C_SDA);                                 // Port-Pin von SDA als Eingang definieren; High-Pegel; SDA = 1 ; durch Pull-Up

    return data;
}


/* -------------------------------------------------------------------------------------
   Funktion :           Mit diese Funktion wird ein Daten-Byte ber die Leitungen des I2C-Bus gesendeten.
                        Dabei erzeugt diese Funktion die Pegelzusnde die fr das jeweilig versendete
                        Daten-Bit bentigte werden. Dabei muss das MSB als erstes Bit versendet werden
   Parameterbergabe:   Input     : I2C_Data   Byte das ber den I2C-Bus gesendet werden soll
                        Return    : -
--------------------------------------------------------------------------------------*/
void I2C_Transmit(uint8_t I2C_Data) 
{

    uint8_t I2C_Daten_Bit_Zaehle=0b10000000;

    while(I2C_Daten_Bit_Zaehle>0)
    {

        DDRB  |= (1<<I2C_SCL);                              // Port-Pin von SCL als Ausgang definieren; Low-Pegel;  SCL = 0;

        if(I2C_Data & I2C_Daten_Bit_Zaehle)
            DDRB  &= ~(1<<I2C_SDA);                         // Port-Pin von SDA als Eingang definieren; High-Pegel; SDA = 1 ; durch Pull-Up
        else
            DDRB  |= (1<<I2C_SDA);                          // Port-Pin von SDA als Ausgang definieren; Low-Pegel;  SDA = 0;

        DDRB  &= ~(1<<I2C_SCL);                             // Port-Pin von SCL als Eingang definieren; High-Pegel; SCL = 1 ; durch Pull-Up

        I2C_Daten_Bit_Zaehle=(I2C_Daten_Bit_Zaehle>>1);
    }


    DDRB  |= (1<<I2C_SCL);                                  // Port-Pin von SCL als Ausgang definieren; Low-Pegel;  SCL = 0;
    DDRB  &= ~(1<<I2C_SDA);                                 // Port-Pin von SDA als Eingang definieren; High-Pegel; SDA = 1 ; durch Pull-Up

    return;
}


/* -------------------------------------------------------------------------------------
   Funktion :           Diese Funktion wartet auf das Acknowledge-Signal des I2C-Slave-Bausteins
                        indem diese Funktion die fr den Empfang des Acknowledge-Signal bentigten
                        Pegelzustnbde auf den I2C-Busleitungen erzeugt, wird der Slave-Baustein dazu
                        aufgefordert die von ihm zuvor empfangenen Daten durch ein Acknowlege-Signal
                        durch ihn zu besttigen.
   Parameterbergabe:   Input     : -
                        Return    : Wert 0 wird zurckgegeben wenn das Acknowledge-Signal empfangen wird
                                    Wert 1 (Fehlerfall) wird zurckgegeben wenn kein Acknowledge-Signal empfangen wird
--------------------------------------------------------------------------------------*/ 
uint8_t I2C_Ack(void) 
{

    uint8_t Time_Out_Zaehler=0;


    while (PINB & (1<<I2C_SDA))                             // warten bist das Ack-Signal besttigt oder Time-Out erreicht wird 
    {
        if(Time_Out_Zaehler<255)
            Time_Out_Zaehler++;
        else
            return 1;
    }

    DDRB  &= ~(1<<I2C_SCL);                                 // Port-Pin von SCL als Eingang definieren; High-Pegel; SCL = 1 ; durch Pull-Up
    asm volatile ("nop");                                    // notwendige Verzgerung
    asm volatile ("nop");                                    // notwendige Verzgerung
    DDRB  |= (1<<I2C_SCL);                                  // Port-Pin von SCL als Ausgang definieren; Low-Pegel;  SCL = 0;

    return 0;
}


/* -------------------------------------------------------------------------------------
   Funktion :           Diese Funktion erzeugt das Acknowledge-Signal des I2C-Master-Bausteins (NAck-Signal)
                        Durch das Setzen der Pegelzustnde der beiden I2C-Leitungen wird in deiser Funktion
   Parameterbergabe:   Input     : -
                        Return    : -
--------------------------------------------------------------------------------------*/ 
void I2C_NAck(void)
{
    DDRB  &= ~(1<<I2C_SCL);                                 // Port-Pin von SCL als Eingang definieren; High-Pegel; SCL = 1 ; durch Pull-Up
    asm volatile ("nop");                                    // notwendige Verzgerung
    asm volatile ("nop");                                    // notwendige Verzgerung
    DDRB  |= (1<<I2C_SCL);                                  // Port-Pin von SCL als Ausgang definieren; Low-Pegel;  SCL = 0;

    return;
}


/* -------------------------------------------------------------------------------------
   Funktion :           Mit Hilfe dieser Funktion werden die fr die Startbedingung des I2C-Buses
                        abentigten Pegelzustnde auf den beiden Leitungen SDA und SCL erzeugt.
   Parameterbergabe:   Input     : -
                        Return    : -
--------------------------------------------------------------------------------------*/ 
void I2C_Stop(void)
{
    DDRB  |= (1<<I2C_SDA);                                  // Port-Pin von SDA als Ausgang definieren; Low-Pegel;  SDA = 0;
    asm volatile ("nop");                                    // notwendige Verzgerung
    asm volatile ("nop");                                    // notwendige Verzgerung
    DDRB  &= ~(1<<I2C_SCL);                                 // Port-Pin von SCL als Eingang definieren; High-Pegel; SCL = 1
    asm volatile ("nop");                                    // notwendige Verzgerung
    asm volatile ("nop");                                    // notwendige Verzgerung
    DDRB  &= ~(1<<I2C_SDA);                                 // Port-Pin von SDA als Eingang definieren; High-Pegel; SDA = 1

    return;
}






/* -------------------------------------------------------------------------------------
   Funktion :           In dieser Funktion verden alle Funktionen zusammengefasst, die fr das
                        Lesen von Daten via I2C-Bus bentigt werden.
   Parameterbergabe:   Input     : Device_Addr    Adresse des Bausteins von dem Daten gelesen werden sollen
                        Return    : data           Eingelesende Daten
--------------------------------------------------------------------------------------*/ 
uint8_t Read_I2C(uint8_t Device_Addr)
{
    uint8_t data=0;

    I2C_Start();                                             // Erzeugt die Start-Bedingung fr den I2C-Bus
    I2C_Transmit(Device_Addr+I2C_READ_BIT);                  // Sendet die Device-Adresse und das Read-Bit an alle Bausteine
    I2C_Ack();                                               // Wartet auf das Acknowledge-Signal des adressierten Bausteines
    data = I2C_Receive();                                    // Liest die Daten vom Bausteins ein
    I2C_NAck();                                              // Erzeugt des Master-Acknowledge-Signal
    I2C_Stop();                                              // Erzeugt die Stopp-Bedingung fr den I2C-Bus

    return data;
}


/* -------------------------------------------------------------------------------------
   Funktion :           In dieser Funktion verden alle Funktionen zusammengefasst, die fr das
                        Schreiben von Daten via I2C-Bus bentigt werden.
   Parameterbergabe:   Input     : Device_Addr    Adresse des Bausteins an den die Daten gesendet werden soll
                                    DataI2C        Daten die an den Baustein gesendet werden sollen
                        Return    : -
--------------------------------------------------------------------------------------*/
void Write_I2C(uint8_t Device_Addr, uint8_t DataI2C)
{

    I2C_Start();                                             // Erzeugt die Start-Bedingung fr den I2C-Bus
    I2C_Transmit(Device_Addr+I2C_WRITE_BIT);                 // Sendet die Device-Adresse und das Write-Bit an alle Bausteine
    I2C_Ack();                                               // Wartet auf das Acknowledge-Signal des adressierten Bausteines
    I2C_Transmit(DataI2C);                                   // Sendet die bergebenen Daten
    I2C_Ack();                                               // Wartet auf das Acknowledge-Signal des adressierten Bausteines
    I2C_Stop();                                              // Erzeugt die Stopp-Bedingung fr den I2C-Bus

    return;
}

