/*
/ Projektname : LCD-I2C-Modul
/ Dateiname   : LCD-I2C-Modul.c
/ Version     : 1.0.0
/ Datum       : 18.01.2014
/ Compiler    : WinAVR 20100110 libc 1_7_1
/ Autor       : Johannes Kretzmann
/ Copyright   : johannes.kretzmann@projektsammlung.de
/               www.projektsammlung.de
/
/
/ Mikrocontroller:                ATtiny2313
/                           ---------------------
/                          |                     |
/                <---PA2---|1(/RESET/dW)   Vcc 20|
/             RX <---PD0---|2(RXD)   (UCSK/SCL)19|---PB7--> ISP Pro.
/             TX <---PD1---|3(TXD)    (MISO/DO)18|---PB6--> ISP Pro.
/                <---PA1---|4(XTAL2) (./DI/SDA)17|---PB5--> ISP Pro.
/                <---PA0---|5(XTAL1)     (OC1B)16|---PB4-->
/        TASTER1 <---PD2---|6(CKOUT/XCK) (OC1A)15|---PB3-->
/        TASTER2 <---PD3---|7(INT1)      (OC0A)14|---PB2-->
/        TASTER3 <---PD4---|8(T0)        (AIN1)13|---PB1--> I2C_SCL
/                <---PD5---|9(OC0B/T1)   (AIN0)12|---PB0--> I2C_SDA
/                          |10 GND        (ICP)11|---PD6-->
/                          |                     |
/                           ---------------------
/
/
/ Remote 8-Bit I/O Expander:       PCF8574
/                               -------------
/                              |             |
/             Jumper JP1  <----|1 A0   Vcc 16|
/             Jumper JP2  <----|2 A1   SDA 15|-----> C PB0
/             Jumper JP3  <----|3 A2   SCL 14|-----> C PB1
/             DISPLAY_DB4 <----|4 P0  /INT 13|----->
/             DISPLAY_DB5 <----|5 P1    P7 12|-----> DISPLAY_BACKLIGHT
/             DISPLAY_DB6 <----|6 P2    P6 11|-----> DISPLAY_E
/             DISPLAY_DB7 <----|7 P3    P5 10|-----> DISPLAY_RW
/                              |8 GND   P4  9|-----> DISPLAY_RS
/                              |             |
/                               -------------
/
/
/ Fuse-Bits:  --- fr internen 8 MHz RC Oszillator und 1 MHz Taktfreqzenz ---
/             -- (1 bedeutet unprogrammiert) --
/             CKSEL3..0   0100           [0100] = Kalibrierter interner RC Oszillator 8 MHz
/                                        [1111] = Externer Quarz/Keramik Resonator von 8 MHz oder grer
/             SUT1..0     11             Quarz Oszillator, mit langsam ansteigender Spannung
/             CKDIV8      0              Teilt den internen RC Oszillator von 8 MHz durch 8 whrend des Start-Up
/
/             --- PonyProg-Setting ---
/             CKSEL3..0   [ ][x][ ][ ]   [ x   ] = Kalibrierter interner RC Oszillator 8 MHz
/                         [ ][ ][ ][ ]   [     ] = Externer Quarz/keramik Resonator von 8 MHz oder grer
/             SUT1..0     [ ][ ]         Quarz Oszillator, mit langsam ansteigender Spannung
/             CKDIV8      [x]            Teilt den internen RC Oszillator von 8 MHz durch 8 whrend des Start-Up
/
/
/ Frequenz:   1.000.000 Hz   (Grundeinstellung fr die Taktfrequenz der CPU bei Werksauslieferung)
/
/
/ Display:    Die Ansteuerung des HD44780 kompatiblen Display-Controllers erfolgt im
/             4-Bit-Interfacemodus. Dabei sind folgende Port-Pins des I2C-Port-Expanders
/             PCF8574 C mit den Pins des Display-Moduls verfunden.
/             P0->DB4; P1->DB5; P2->DB6; P3->DB7; P4->RS; P5->RW, P6->E; P7->BACKLIGHT
/
/
/ ATtiny2313 
/ Ports:      Taster:  PD2 -> TASTER1
/                      PD3 -> TASTER2
/                      PD4 -> TASTER3
/             I2C   :  PB0 -> I2C_SDA
/                      PB1 -> I2C_SCL
/
/
/ PCF8574 
/ Ports:      Jumper:  A0  -> JP1
/                      A1  -> JP2
/                      A2  -> JP3
/             Display: P0  -> DB4
/                      P1  -> DB5
/                      P2  -> DB6
/                      P3  -> DB7
/                      P4  -> RS
/                      P5  -> RW
/                      P6  -> E
/                      P7  -> BACKLIGHT
/
/
/
/
/
/               -------------------------------- Beschreibungen --------------------------------------
/ LCD-I2C:      Dieses Programm ist von mir dazu geschrieben worden um ein LC-Display per I2C-Bus anzusteuern,
/               dessen integrierten Controller den Befehlssatz des HD44780-Standart von Hitachi untersttzt.
/               Die Ansteuerung erfolgt mittels des ATMEL ATtiny2313 Mikrocontrollers und den I2C-Port-Expander
/               PCF8574 der Firma Philips an dem das LC-Display angeschlossen ist.
/               Da der ATtiny2312 von Haus aus keine I2C-Bus-Schnittstelle besitzt, wird diese per Software durch
/               dieses Programm implementiert.
/               Die von mir zur Entwicklung dieses Programmes eingesetzten Komponenten habe ich bei der Firma
/               Pollin Electronic gekauft. Diese knnen unter folgender Bezeichnung und Bestellnummer auf der Homepage
/               www.pollin.de bezogen werden. Zuerst wird das ATMEL Evaluations-Board als Bausatz Artikel-Nr. 810 030
/               bzw. Fertigmodul Artikel-Nr. 810 074 bentigt, auf dem sich der Mikrocontroller und die Programmier-
/               Schnittstelle ISP befindet. Als LC-Displays knnen unter anderem das 2-zeilige mit 16 Zeichen je Zeile
/               und gelber Hintergrundbeleuchtung mit der Artikel-Nr. 120 420 bzw. mit blauer Hintergrundbeleuchtung
/               Artikel-Nr. 120 421 verwendet werden. Die LC-Displays knnen direkt auf das LCD-I2C-Modul mit der 
/               Artikel-Nr. 810 145 aufgesteckt werden, auf dem sich auch der Baustein PCF8574 also der I2C-Port-Expander
/               befindet. Zur leichteren Verdrahtung der Baustze habe ich noch den Bausatz Portsplitter Artikel-Nr.
/               810 161 sowie das IDE-Kabel Artikel-Nr. 720 296 verwendet.
/               Die Tasten 1, 2 und 3 die sich auf der Evaluations-Board befinden werden wie folgend beschrieben in dem
/               Programm eingesetzt. Durch drcken der Taste 1 wird der Text der hinter den beiden Konstanten Display_Text1
/               und Display_Text2 hinterlegt ist am LC-Display angezeigt. Mit Drcken der Taste 2 wird der Text der
/               Konstanten Display_Text3 und Display_Text4 ausgegeben. Die Taste 2 schaltet die Hintergrundbeleuchtung
/               des LC-Displays an bzw. aus.
/               Mittels der seriellen Schnittstelle die sich auch auf dem Evaluations-Board befindet, werden beim 
/               bettigen der Tasten die Texte sowie der Status des I2C-Buses ausgegeben.
/
/ Achtung:      Bitte beachten Sie, dass dieses Programm nur uneingeschrnkt fr nicht kommerzielle Zwecke genutzt
/               werden darf. Ich bernehme keine Garantie fr die einwandfreie Funktion dieses Programmes und ber-
/               nehme dadurch auch keine Haftung fr Folgeschden die Aufgrund der Benutzung dieser Daten entstehen.
/               Dieses gilt auch dann, wenn ich ber einen solchen Schaden unterrichtet worden bin.
/
/
/ Zusammenfassung:       Taster 1:    Ausgabe Text 1 am LC-Display
/                        Taster 2:    Ausgabe Text 2 am LC-Display
/                        Taster 3:    Ein- bzw. Ausschalten der LC-Display-Hintergrundbeleuchtung
/
/
*/


#include <string.h>
#include <avr/io.h>
#include <inttypes.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include "I2C-Display.h"
#include "I2C.h"



#define TASTER1                 2             // Port PD2 - meinSpezialByte
#define TASTER2                 3             // Port PD3 - meinSpezialByte
#define TASTER3                 4             // Port PD4 - meinSpezialByte

#define BACKLIGHT               1             // meinSpezialByte


#define F_CPU                   1000000UL     // Systemtakt 1 MHz
#define UART_BAUD_RATE          9600UL        // UART-Baut-Rate
#define QBLength_TX             20            // TX-Pufferspeichergre

#define BASE_BIN                2
#define BASE_OCT                8
#define BASE_DEZ                10
#define BASE_HEX                16


#define STRING_LAENGE_utoa      4
#define STRING_LENGTH_MAX       25

#define SERIELL_TRACKING


/* -------------------------------------------------------------------------------------
   Definition der im Programm verwendeten Displaytextanzeigen
--------------------------------------------------------------------------------------*/
#define DISPLAY_TEXT0           0
#define DISPLAY_TEXT1           1
#define DISPLAY_TEXT2           2
#define DISPLAY_TEXT3           3
#define DISPLAY_TEXT4           4
#define DISPLAY_TEXT5           5
#define DISPLAY_TEXT6           6
#define HINWEIS_TEXT1           7
#define HINWEIS_TEXT2           8
#define HINWEIS_TEXT3           9
#define HINWEIS_TEXT4           10
#define DISPLAY_DATA            11
#define DISPLAY_READ_I2C        12



/* -------------------------------------------------------------------------------------
   Festlegung der Texte die auf dem Display angezeigt werden sollen
--------------------------------------------------------------------------------------*/ 
const char Display_Text0[]     PROGMEM = " LCD-I2C Modul  ";
const char Display_Text1[]     PROGMEM = "www.projekt-    ";
const char Display_Text2[]     PROGMEM = "     sammlung.de";
const char Display_Text3[]     PROGMEM = "Johannes        ";
const char Display_Text4[]     PROGMEM = "Kretzmann       ";
const char Display_Text5[]     PROGMEM = " Backlight an   ";
const char Display_Text6[]     PROGMEM = " Backlight aus  ";
const char Hinweis_Text1[]     PROGMEM = " Adressierung des ";
const char Hinweis_Text2[]     PROGMEM = "I2C-Bausteins ";
const char Hinweis_Text3[]     PROGMEM = "ist fehlgeschlagen";
const char Hinweis_Text4[]     PROGMEM = " Fehlerhafte I2C-Leitung";

const char Display_Data[]      PROGMEM = "\n\rData :";
const char Display_Read_I2C[]  PROGMEM = "\n\rRead_I2C :";



const char *Display_Text[] PROGMEM = { Display_Text0, Display_Text1, Display_Text2,
                                        Display_Text3, Display_Text4, Display_Text5,
                                        Display_Text6, 
                                        Hinweis_Text1, Hinweis_Text2, Hinweis_Text3,
                                        Hinweis_Text4,
                                        Display_Data,  Display_Read_I2C};




/* -------------------------------------------------------------------------------------
   Deklaration und Wertefestlegung der globalen Variablen
--------------------------------------------------------------------------------------*/
volatile uint8_t  TXQueueBuffer[QBLength_TX];                           // Zwischenspeicher fr zu sendende Daten (volatile steht fr Variablen
volatile uint8_t  pInQueueTX  = 0;                                      // die in Interrupt-Funktionen manipuliert werden knnen)
volatile uint8_t  pOutQueueTX = 0;




/* -------------------------------------------------------------------------------------
   USART-TX-Interrupt-Funktion:
   Parameterbergabe: Parameter werden in globalen Variablen bergeben. 
--------------------------------------------------------------------------------------*/
ISR(USART_UDRE_vect)
{
    uint8_t temp_sreg_TX;

    temp_sreg_TX = SREG;                                                // Status-Register sichern

    UDR=TXQueueBuffer[pOutQueueTX];	                                    // Sendet Zeichen

    if(pOutQueueTX == QBLength_TX-1)
       pOutQueueTX = 0;
    else
       pOutQueueTX++;


    if(pOutQueueTX == pInQueueTX)
    {
       UCSRB   &= ~(1<<UDRIE);                                          // UART TX Interrupt fr leeres Sendedatenregister ausschalten
    }

    SREG = temp_sreg_TX;                                                // Status-Register rckschreiben

    return;
}



/* -------------------------------------------------------------------------------------
   Funktion :           PutStringInQueueTX()
   Parameterbergabe:   Input     : Zeichenkette
                        Return    : -
--------------------------------------------------------------------------------------*/
void PutStringInQueueTX(char *Zeichenkette)
{
    uint8_t  i;

    for(i=0; Zeichenkette[i] != 0; i++)
    {
       // Zwangspause fr den -Controller falls dieser schneller ist als die UART
    while((pInQueueTX+1 == pOutQueueTX) || ((pInQueueTX+1 == QBLength_TX-1) && (pOutQueueTX == 0)))
    { }
    
       TXQueueBuffer[pInQueueTX] = Zeichenkette[i];

       if(pInQueueTX == QBLength_TX-1)
          pInQueueTX = 0;
       else
          pInQueueTX++;
    }

    UCSRB |= (1<<UDRIE);                                      // UART TX Interrupt fr leeres Sendedatenregister einschalten

    return; 
}



/* -------------------------------------------------------------------------------------
   Funktion :           PutDataInQueueTX()
   Parameterbergabe:   Input     : data
                        Return    : -
--------------------------------------------------------------------------------------*/
void PutDataInQueueTX(uint8_t data)
{
    // Zwangspause fr den -Controller falls dieser schneller ist als die UART
    while((pInQueueTX+1 == pOutQueueTX) || ((pInQueueTX+1 == QBLength_TX-1) && (pOutQueueTX == 0)))  {}

       TXQueueBuffer[pInQueueTX] = data;

       if(pInQueueTX == QBLength_TX-1)
          pInQueueTX = 0;
       else
          pInQueueTX++;

    UCSRB |= (1<<UDRIE);                                      // UART TX Interrupt fr leeres Sendedatenregister einschalten
 
    return; 
} 



/* -------------------------------------------------------------------------------------
   Funktion :           Die Funktion Serielle_Display_Anzeige() liest entsprechend dem bergebenen
                        Parameter den ihm zugeteilten Text aus dem Flash-Speicher aus und bergibt
                        diesen der Funktion PutStringInQueueTX() durch die der Text dann ber die
                        serielle Schnittstelle bertragen wird.
   Parameterbergabe:   Input     : Text_Nr    bergibt die Nummer des zu anzeigenden Textes
                        Return    : -
--------------------------------------------------------------------------------------*/
void Serielle_Display_Anzeige(uint8_t Text_Nr)
{
    char meinHilfsString[STRING_LENGTH_MAX];

    strcpy_P(meinHilfsString,  (const char*) (pgm_read_word(&(Display_Text[Text_Nr]))));
    PutStringInQueueTX(meinHilfsString);

    return;
}
  


/* -------------------------------------------------------------------------------------
   Funktion :           Die Funktion Display_Anzeige() liest entsprechend dem bergebenen Parameter
                        den ihm zugeteilten Text aus dem Flash-Speicher aus und bergibt diesen
                        Text der Funktion WriteDisplayString() durch die der Text dann am Display
                        angezeigt wird. 
   Parameterbergabe:   Input     : Text_Nr    bergibt die Nummer des zu anzeigenden Textes
                        Return    : -
--------------------------------------------------------------------------------------*/
void Display_Anzeige(uint8_t Text_Nr)
{
    char meinHilfsString[STRING_LENGTH_MAX];

    strcpy_P(meinHilfsString,  (const char*) (pgm_read_word(&(Display_Text[Text_Nr]))));
    WriteDisplayString(meinHilfsString);

    return;
}



/* -------------------------------------------------------------------------------------
   Hauptprogramm    :   LCD-I2C-Modul (Beschreibung siehe Einleitung)
   Parameterbergabe:   Input     : -
                        Return    : -
--------------------------------------------------------------------------------------*/ 
int main(void)
{
    // ================= Initialisierung ===================

    // ---------- Portpinbelegung und Definition -----------

    // ------ Port D (Taster1/2 und 3) ------
    DDRD  &= ~((1<<TASTER1) | (1<<TASTER2) | (1<<TASTER3));        // definiert Port-Pins fr TASTER als Eingnge
    PORTD &= ~((1<<TASTER1) | (1<<TASTER2) | (1<<TASTER3));        // deaktiviere internen Pull-Ups von Port-Pin

    // -------------------- UART --------------------
    uint16_t  myUBRR;                                                 // 16-Bit breite Speichervariable fr Bautratenberechnung 
    // myUBRR  = F_CPU/(UART_BAUD_RATE*16L)-1;                        // Berechnung der Bautratenwerte fr das UBRR-Register im Asynchronous Normal mode (U2X=0)
    myUBRR  = F_CPU/(UART_BAUD_RATE*8L)-1;                           // Berechnung der Bautratenwerte fr das UBRR-Register im Asynchronous Double Speed mode (U2X=1)
    UCSRA   = (1<<U2X);                                              // Asynchronous Double Speed mode aktivieren
    UBRRH   = (uint8_t) (myUBRR>>8);                                 // Setzt Highbyte des UBRR-Registers
    UBRRL   = (uint8_t) myUBRR;                                      // Setzt Lowbyte  des UBRR-Registers
    UCSRB   = (1<<TXEN);                                             // UART TX einschalten
    UCSRC   = (1<<UCSZ1) | (1<<UCSZ0);                              // Setzt die Bits fr die Betriebsart Asynchron 8N1

    sei();                                                           // aktiviert global alle definierten Interrupte

    Init_I2C();
    InitDisplay();




    // ------------ lokale Variablen ------------
    uint8_t FehlerCode;                                              // Temporrer Speicher fr bettigte Taster
    uint8_t meinSpezialByte;                                         // Temporrer Speicher fr bettigte Taster

    meinSpezialByte = (0b00000000) | (1<<BACKLIGHT);

    // Prfung ob I2C-Leitungen fehlerhaft sind
    if (PINB & (1<<I2C_SDA))                                        // an den Leitungen sollte ein High-Pegel anliegen
    {
        // Prfung ob I2C-Bus-Verbindung zum Display besteht
        I2C_Start();                                                // Erzeugt die Start-Bedingung fr den I2C-Bus
        I2C_Transmit(DEVICE_ADDR+I2C_READ_BIT);                     // Sendet die Device-Adresse und das Read-Bit an alle Bausteine
        FehlerCode=I2C_Ack();                                       // Wartet auf das Acknowledge-Signal des adressierten Bausteins
        I2C_Stop();                                                 // Erzeugt die Stopp-Bedingung fr den I2C-Bus
    }
    else 
        FehlerCode=2;



    #ifdef SERIELL_TRACKING

        PutStringInQueueTX("\n\n\r");
        Serielle_Display_Anzeige(DISPLAY_TEXT0);
        PutStringInQueueTX("\n\n\r");

        if(FehlerCode==0 || FehlerCode==1)
        {
            Serielle_Display_Anzeige(HINWEIS_TEXT1);
            Serielle_Display_Anzeige(HINWEIS_TEXT2);
        }
        if(FehlerCode==1)
            Serielle_Display_Anzeige(HINWEIS_TEXT3);

        if(FehlerCode==2)
            Serielle_Display_Anzeige(HINWEIS_TEXT4);
         
    #endif

    SetDisplayCursor(1,0);
    Display_Anzeige(DISPLAY_TEXT0);


    while(1)
    {


        // ====== Taster 1 ======
        if(PIND & (1<<TASTER1))
        {
           if(~meinSpezialByte & (1<<TASTER1))        // Beim ersten Tastendruck wahr
           {   meinSpezialByte |= (1<<TASTER1);       // setzt Bit fr Taster1
 
               #ifdef SERIELL_TRACKING
               PutStringInQueueTX("\n\r");
               Serielle_Display_Anzeige(DISPLAY_TEXT1);
               PutStringInQueueTX("\n\r");
               Serielle_Display_Anzeige(DISPLAY_TEXT2);
               #endif

               SetDisplayCursor(1,0);
               Display_Anzeige(DISPLAY_TEXT1);
               SetDisplayCursor(2,0);
               Display_Anzeige(DISPLAY_TEXT2);
            }
        }
        else
           meinSpezialByte &= ~(1<<TASTER1);         // lscht Bit fr Taster1


        // ====== Taster 2 ======
        if(PIND & (1<<TASTER2))
        {
           if(~meinSpezialByte & (1<<TASTER2))        // Beim ersten Tastendruck wahr
           {   meinSpezialByte |= (1<<TASTER2);       // setzt Bit fr Taster2

               #ifdef SERIELL_TRACKING
               PutStringInQueueTX("\n\r");
               Serielle_Display_Anzeige(DISPLAY_TEXT3);
               PutStringInQueueTX("\n\r");
               Serielle_Display_Anzeige(DISPLAY_TEXT4);
               #endif

               SetDisplayCursor(1,0);
               Display_Anzeige(DISPLAY_TEXT3);
               SetDisplayCursor(2,0);
               Display_Anzeige(DISPLAY_TEXT4);
            }
        }
        else
           meinSpezialByte &= ~(1<<TASTER2);         // lscht Bit fr Taster2


        // ====== Taster 3 ======
        if(PIND & (1<<TASTER3))
        {
           if(~meinSpezialByte & (1<<TASTER3))        // Beim ersten Tastendruck wahr
           {   meinSpezialByte |= (1<<TASTER3);       // setzt Bit fr Taster3

                SetDisplayCursor(1,0);
                Display_Anzeige(DISPLAY_TEXT0);

                if(meinSpezialByte & (1<<BACKLIGHT))
                {
                    meinSpezialByte &= ~(1<<BACKLIGHT);

                    SetDisplayCursor(2,0);
                    Display_Anzeige(DISPLAY_TEXT6);

                    // Setze Bit fr Hintergrundbeleuchtung auf aus
                    PCF8574_Port_Byte |= (1<<I2C_PORT_BIT_BACKLIGHT);
                    Write_I2C(DEVICE_ADDR, PCF8574_Port_Byte);

                    #ifdef SERIELL_TRACKING
                    PutStringInQueueTX("\n\r");
                    Serielle_Display_Anzeige(DISPLAY_TEXT6);
                    #endif
                }
                else
                {
                    meinSpezialByte |= (1<<BACKLIGHT);

                    SetDisplayCursor(2,0);
                    Display_Anzeige(DISPLAY_TEXT5);

                    // Setze Bit fr Hintergrundbeleuchtung auf ein
                    PCF8574_Port_Byte &= ~(1<<I2C_PORT_BIT_BACKLIGHT);
                    Write_I2C(DEVICE_ADDR, PCF8574_Port_Byte);

                    #ifdef SERIELL_TRACKING
                    PutStringInQueueTX("\n\r");
                    Serielle_Display_Anzeige(DISPLAY_TEXT5);
                    #endif
                }

            }
        }
        else
           meinSpezialByte &= ~(1<<TASTER3);         // lscht Bit fr Taster3

    }

}


